#include	<stdio.h>
#include	<math.h>
#include	<X11/Intrinsic.h>
#include	<X11/XWDFile.h>
#include	"gene.h"
#include	"draw.h"
#define	MaxObjects	0x20000
#define	GAMMA	0.6

static	int	Floor1RGB[3] = { 0xf0, 0xd0, 0x70 },
		Floor2RGB[3] = { 0xe0, 0xa0, 0xd0 },
		SkyRGB[3] = { 0x80, 0xf0, 0xff },
		SoilRGB[3] = { 0x80, 0x40, 0x10 };
static	int	NColumns,IndWidth,IndHeight;
static	int	NObject;
static	Cell	*ObjectBuffer[MaxObjects];
static	double	vx = 0, vy = 0, vz = 5.0;	/* view position */
static	double	sinrx,cosrx,sinry,cosry,sinrz,cosrz;	/* sin & cos of vr */
static	double	scale;
static	int	CenterX, CenterY;
static	unsigned char	ZoomMode = 0;
static	Individual	*ZoomedIndividual;

extern	int	PopulationSize;
extern	Individual	*Population;
extern	unsigned long	Step;
extern	double	CellSize;
extern	int	WindowWidth, WindowHeight;
extern	Display	*dpy;
extern	Window	window;
extern	Pixmap	pixmap;
extern	int	depth;
extern	GC	gc;
extern	Colormap	cmap;
extern	Visual	visual;
extern	unsigned long	fg, bg;
extern	int	IScale;
extern	Boolean	BlackBackground;
extern	DrawingShape	DrawShape;
extern	long	DrawPopSize;
extern	double	RotateX, RotateY, RotateZ, ScaleR;
extern	double	ViewDistance, ViewShift, ViewHeight;

#define	COLOR_GRADES	8
#define	NUT_GRADES	5
#define	NCELLSTATES	(NUT_GRADES*NUT_GRADES)
#define	NCOLORS		(COLOR_GRADES*NCELLSTATES)
static	struct	{ int red, green, blue; } CellColors[NCELLSTATES];
unsigned long	pixel[NCOLORS*2];
double	radius_rate[COLOR_GRADES];

static void pseudo_color_init()
{
	XColor	acolor[NCOLORS], *cp;
	unsigned long	pl[1];
	int	i, j, k;
	
	if (!XAllocColorCells(dpy, cmap, True, pl, 0, pixel, NCOLORS)) {
fprintf(stderr,"Sorry, there is no enough room to allocate color maps.\n");
		exit(1);
	}
	memcpy(pixel+NCOLORS,pixel,sizeof(long)*NCOLORS);
	for (i = k = 0, cp = acolor; i < NCELLSTATES; i ++) {
		for (j = 1; j <= COLOR_GRADES; j ++, k ++, cp ++) {
		cp->pixel = pixel[k];
		cp->flags = DoRed | DoGreen | DoBlue;
		cp->red = (CellColors[i].red * j * 255)/ COLOR_GRADES;
		cp->green = (CellColors[i].green * j * 255) / COLOR_GRADES;
		cp->blue = (CellColors[i].blue * j * 255) / COLOR_GRADES;
		}
	}
	XStoreColors(dpy, cmap, acolor, NCOLORS);
}
static	int	red_shift, green_shift, blue_shift;
static	double	FloorGrid;
static	unsigned long	shift_mask(c,s,m) int c,s; unsigned long m;
{
	return (s > 0)? (unsigned long)(c << s) & m
	 : (unsigned long)(c >> -s) & m;
}
unsigned long rgb_to_pixel(r,g,b) int r, g, b;
{
	return shift_mask(r,red_shift,visual.red_mask)
	| shift_mask(g,green_shift,visual.green_mask)
	| shift_mask(b,blue_shift,visual.blue_mask);
}
static void true_color_init()
{
	int	i, j, k, px;
	for (i = 0, k = 1; i < depth; i ++, k <<= 1)
		if (k & visual.red_mask) red_shift = i - 7;
	for (i = 0, k = 1; i < depth; i ++, k <<= 1)
		if (k & visual.green_mask) green_shift = i - 7;
	for (i = 0, k = 1; i < depth; i ++, k <<= 1)
		if (k & visual.blue_mask) blue_shift = i - 7;
	for (i = k = 0; i < NCELLSTATES; i ++)
	for (j = 1; j <= COLOR_GRADES; j ++, k ++) {
		pixel[k] = rgb_to_pixel(CellColors[i].red*j/COLOR_GRADES,
			CellColors[i].green*j/COLOR_GRADES,
			CellColors[i].blue*j/COLOR_GRADES);
		pixel[k+NCOLORS] = rgb_to_pixel(
			(CellColors[i].red*j/2+128)/COLOR_GRADES,
			(CellColors[i].green*j/2+128)/COLOR_GRADES,
			(CellColors[i].blue*j/2+128)/COLOR_GRADES);
	}
	FloorGrid = CellSize * 4.;
}
draw_init()
{
	int	i, j, k;
	for (i = k = 0; i < NUT_GRADES; i ++)
	for (j = 0; j < NUT_GRADES; j ++, k ++) {
		CellColors[k].red = 255 - (i+j)*64/NUT_GRADES;
		CellColors[k].green = (i + 1) * (256/NUT_GRADES) - 1;
		CellColors[k].blue = (j + 1) * (256/NUT_GRADES) - 1;
	}
	switch (visual.class) {
		case PseudoColor: pseudo_color_init(); break;
		case TrueColor: true_color_init(); break;
		default: fprintf(stderr,
		"Sorry, visual class #%d is not supported.\n",visual.class);
		exit(1);
	}
	for (i = 0; i < COLOR_GRADES; i ++)
		radius_rate[i] = sin((COLOR_GRADES-i)*M_PI/(2*COLOR_GRADES));
}
static	void	set_view_pose(c,dx,dy,dz,dr)
Cell	*c;
double	*dx, *dy, *dz, *dr;
{
	extern	double	ThGlucose, ThH2O;
	double	x, y, z, r, p, q, zz;
	Cell	*m = c->mother;
	x = c->x; y = c->y; z = c->z; r = CellSize;
	if (m && c->rr < 1.
	 && (DrawShape == DrSpGrowingSpheres || DrawShape == DrSpDeluxe)) {
		p = c->glu / ThGlucose; q = c->h2o / ThH2O;
		zz = (p < q)? p : q; if (zz > 1.) zz = 1.;
		if (zz > c->rr) c->rr = zz;
		if (c->rr < 1.) {
			p = (1. - c->rr) / 2.;
			x -= (x - m->x) * p;
			y -= (y - m->y) * p;
			z -= (z - m->z) * p;
			r *= c->rr;
		}
	}
	p = y * cosrz - x * sinrz;
	q = y * sinrz + x * cosrz;
	y = p; x = q;
	p = x * cosry - z * sinry;
	q = x * sinry + z * cosry;
	x = p; z = q;
	p = z * cosrx - y * sinrx;
	q = z * sinrx + y * cosrx;
	z = p; y = q;
	*dx = x + vx + ViewShift;
	*dy = y + vy + ViewHeight;
	*dz = z + vz + ViewDistance;
	*dr = r;
}

#define	ViewArrayWidth	128
#define	ViewArraySize	(ViewArrayWidth*ViewArrayWidth)
#define	ViewArrayUnit	7
typedef	struct varray_rec {
	struct varray_rec	*next;
	Cell	*cell[ViewArrayUnit];
} ViewArrayElm;
static	ViewArrayElm	**ViewArray = NULL;
static	Cell	*DrawCell;
static	double	Sx, Sy, AA, HitZ;
static	double	VPx, VPy, VPz;	/* position of view point */
static	int	check_draw_cell(c,hz) Cell *c; double *hz; {
	double	b = Sx * c->draw_x + Sy * c->draw_y + c->draw_z;
	double	d = b * b - AA * c->C;
	if (d >= 0.) {
		*hz = (b - sqrt(d)) / AA;
		return (*hz > 0.);
	} else	return 0;
}
static	void	set_bond_position(c,m,b) Cell *c, *m, *b; {
	b->draw_x = (c->draw_x + m->draw_x) * .5;
	b->draw_y = (c->draw_y + m->draw_y) * .5;
	b->draw_z = (c->draw_z + m->draw_z) * .5;
	b->draw_r = hypot(c->draw_r, m->draw_r) * .5;
	b->C = b->draw_x * b->draw_x + b->draw_y * b->draw_y
		+ b->draw_z * b->draw_z - b->draw_r * b->draw_r;
}
static	double	bond_color_rate(c,x,y,z) Cell *c; double x,y,z; {
	double	dx = c->draw_x - x, dy = c->draw_y - y, dz = c->draw_z - z;
	return sqrt(dx*dx + dy*dy + dz*dz) - c->draw_r;
}
static	void	find_draw_cell(c) Cell *c;
{
	static	Cell	bond;
	double	hzc, hzb;
	Cell	*m, b;
	int	tb, tc = check_draw_cell(c,&hzc);
	if (m = c->mother) {
		set_bond_position(c,m,&b);
		tb = check_draw_cell(&b,&hzb);
	} else	tb = 0;
	if (!tc && !tb) return;
	if ((!tc && tb) || (tb && hzc > hzb)) { hzc = hzb; tb = 2; }
	if (HitZ < hzc) return;
	if (tb == 2) {
		double	hx = Sx * hzb, hy = Sy * hzb;
		double	lc, lm;
		lc = bond_color_rate(c,hx,hy,hzb);
		lm = bond_color_rate(m,hx,hy,hzb);
		b.glu = (c->glu * lm + m->glu * lc) / (lc + lm);
		b.h2o = (c->h2o * lm + m->h2o * lc) / (lc + lm);
		bond = b;
		DrawCell = &bond;
	} else	DrawCell = c;
	HitZ = hzc;
	return;
}
static	double	under_ground_p()
{
	double	y;
	y = ((Sy * HitZ - ViewHeight) * cosrx
	   - (HitZ - vz - ViewDistance) * sinrx) * cosrz + Sx * HitZ * sinrz;
	return (VPy < 0.)? y : -y;
}
static	void	retro_position(x,y,z,u,v,w)
double	x, y, z, *u, *v, *w;
{
	double	p, q;
	p = z * cosrx + y * sinrx;
	q = y * cosrx - z * sinrx;
	z = p; y = q;
	p = x * cosry + z * sinry;
	q = z * cosry - x * sinry;
	x = p; z = q;
	p = y * cosrz + x * sinrz;
	q = x * cosrz - y * sinrz;
	*u = q; *v = p; *w = z;
}
static	void	background_scene(r,g,b) int *r, *g, *b;
{
	double	u, v, w, x, z, dx, dz, d;
	int	*p;
	retro_position(Sx,Sy,1.,&u,&v,&w);
	if (v >= -1e-10 && VPy >= 0.) {
		*r = SkyRGB[0]; *g = SkyRGB[1]; *b = SkyRGB[2];
		return;
	} else if (v <= 1e-10 && VPy < 0.) {
		*r = SoilRGB[0]; *g = SoilRGB[1]; *b = SoilRGB[2];
		return;
	}
	x = VPx - (dx = VPy * u / v);
	z = VPz - (dz = VPy * w / v);
	p = (((int)(x/FloorGrid)^(int)(z/FloorGrid)^(x * z > 0.))&1)?
		Floor1RGB : Floor2RGB;
	d = sqrt(dx*dx+dz*dz+VPy*VPy)*.05 + 1.;
	if (VPy >= 0.) {
		*r = 255-(255-p[0])/d;
		*g = 255-(255-p[1])/d;
		*b = 255-(255-p[2])/d;
	} else { *r = p[0]/d; *g = p[1]/d; *b = p[2]/d; }
}
static int view_array_x(x) int x; {
	int	v = x * ViewArrayWidth / WindowWidth;
	if (v < 0) v = 0; else if (v >= ViewArrayWidth) v = ViewArrayWidth-1;
	return v;
}
static int view_array_y(y) int y; {
	int	v = y * ViewArrayWidth / WindowHeight;
	if (v < 0) v = 0; else if (v >= ViewArrayWidth) v = ViewArrayWidth-1;
	return v;
}
static ViewArrayElm *FreeViewArrayElm = NULL;
static void free_viewarrayelm() {
	int	i;
	ViewArrayElm	*p;
	for (i = 0; i < ViewArraySize; i ++) if (ViewArray[i]) {
		for (p = ViewArray[i]; p->next; p = p->next);
		p->next = FreeViewArrayElm;
		FreeViewArrayElm = ViewArray[i];
		ViewArray[i] = NULL;
	}
}
static ViewArrayElm *new_viewarrayelm(c) Cell *c;
{
	ViewArrayElm	*new = FreeViewArrayElm;
	if (!new) {
		int	i;
		FreeViewArrayElm
			= (ViewArrayElm *)malloc(sizeof(ViewArrayElm)*64);
		if (!FreeViewArrayElm)
			{ perror("malloc for ViewArrayElm"); exit(1); }
		for (i = 0; i < 63; i ++)
			FreeViewArrayElm[i].next = FreeViewArrayElm + i + 1;
		FreeViewArrayElm[63].next = NULL;
		new = FreeViewArrayElm;
	}
	FreeViewArrayElm = new->next;
	memset(new,0,sizeof(ViewArrayElm));
	new->cell[0] = c;
	return new;
}
static void add_view_cell(elm,c) ViewArrayElm **elm; Cell *c;
{
	if (*elm) {
		ViewArrayElm	*last;
		int	i;
		for (last = *elm; last->next; last = last->next);
		for (i = 0; i < ViewArrayUnit; i ++) if (!last->cell[i]) {
			last->cell[i] = c; return;
		}
		elm = &(last->next);
	}
	*elm = new_viewarrayelm(c);
}
static void set_view_array_elm(b,c) Cell *b,*c;
{
	int	x1, y1, x2, y2, i, j, n;
	x1 = view_array_x((int)floor(b->draw_x - b->draw_r));
	x2 = view_array_x((int)ceil(b->draw_x + b->draw_r));
	y1 = view_array_y((int)floor(b->draw_y - b->draw_r));
	y2 = view_array_y((int)ceil(b->draw_y + b->draw_r));
	for (j = y1; j <= y2; j ++) for (i = x1; i <= x2; i ++)
		add_view_cell(ViewArray+j*ViewArrayWidth+i,c);
}	
static void set_view_array(c) Cell *c;
{
	Cell	*m, b;
	set_view_array_elm(c,c);
	if (m = c->mother) {
		set_bond_position(c,m,&b);
		set_view_array_elm(&b,c);
	}
}
#define	TB1	1.
#define	TB2	4.
#define	TRANSPARENT(f,b,d)	((f)+(b)*(TB1+(d)*TB2))/(TB1+1.+(d*TB2))
static	int	calc_rgb(wx,wy,velm,r,g,b)
int	wx, wy,*r,*g,*b;
ViewArrayElm	*velm;
{
	int	i;
	ViewArrayElm	*p;
	DrawCell = NULL; HitZ = 1e30;
	Sx = (wx - CenterX) / scale;
	AA = Sx * Sx + Sy * Sy + 1.;
	for (p = velm; p; p = p->next) for (i = 0; i < ViewArrayUnit; i ++) {
		if (p->cell[i]) find_draw_cell(p->cell[i]);
		else break;
	}
	if (DrawCell) {
		int	br, bg, bb;
		double	dpt;
		double	cr = cos(hypot(Sx*HitZ-DrawCell->draw_x,
			Sy*HitZ-DrawCell->draw_y)/DrawCell->draw_r*M_PI/2);
		*g = pow(DrawCell->glu*cr,GAMMA) * 255;
		*b = pow(DrawCell->h2o*cr,GAMMA) * 255;
		*r = pow((1.-(DrawCell->glu + DrawCell->h2o)/2.)*cr,GAMMA) * 255;
		if ((dpt = under_ground_p()) > 0.) {
			background_scene(&br,&bg,&bb);
			*r = TRANSPARENT(*r,br,dpt);
			*g = TRANSPARENT(*g,bg,dpt);
/*			*g = (*g / 2 + 64 + bg) / 2; */
			*b = TRANSPARENT(*b,bb,dpt);
		}
		return 1;
	} else {
		background_scene(r,g,b);
		return 0;
	}
}
#define	CAP_XOFFSET	20
static	unsigned long	gray = 0;
static void draw_string(gc,str,fst,x,y,align)
GC	gc;
char	*str;
XFontStruct	*fst;
int	x, y, align;
{
	int width = XTextWidth(fst,str,strlen(str));
	switch (align) {
		case 0: x += CAP_XOFFSET; break;
		case 1: x += (WindowWidth - width) / 2; break;
		case 2: x = WindowWidth - x - width - CAP_XOFFSET;
	}
	if (DrawShape == DrSpDeluxe) {
		int b = (fst->max_bounds.ascent + fst->max_bounds.descent) / 14;
		XSetForeground(dpy,gc,gray);
		XDrawString(dpy,pixmap,gc,x+b,y+b,str,strlen(str));
		XDrawString(dpy,window,gc,x+b,y+b,str,strlen(str));
		XSetForeground(dpy,gc,fg);
		XDrawString(dpy,window,gc,x,y,str,strlen(str));
	}
	XDrawString(dpy,pixmap,gc,x,y,str,strlen(str));
}
#define	FMTLEN	32
static void draw_caption(ind,str,fst,x,y,align)
Individual	*ind;
char	*str;
XFontStruct	*fst;
int	x, y, align;
{
	unsigned long	k;
	char	*p, *q, fmt[FMTLEN], buf[128];
	int	i, flg, LineHeight;
	static	GC	cap_gc = NULL;
	extern	unsigned long	Generation;
	extern	XFontStruct	*Gfont;
	if (!cap_gc) {
		XColor	xcol[2];
		cap_gc = XCreateGC(dpy,window,0,0);
		XCopyGC(dpy,gc,(1<<(GCLastBit+1))-1,cap_gc);
		xcol[0].pixel = fg;
		xcol[1].pixel = bg;
		xcol[0].flags = xcol[1].flags = DoRed | DoGreen | DoBlue;
		XQueryColors(dpy,cmap,xcol,2);
		xcol[0].red = ((int)xcol[0].red + (int)xcol[1].red*3)/4;
		xcol[0].green = ((int)xcol[0].green + (int)xcol[1].green*3)/4;
		xcol[0].blue = ((int)xcol[0].blue + (int)xcol[1].blue*3)/4;
		gray = XAllocColor(dpy,cmap,xcol)? xcol[0].pixel : fg;
	}
	if (!fst) fst = Gfont;
	XSetFont(dpy,cap_gc,fst->fid);
	LineHeight = (fst->max_bounds.ascent + fst->max_bounds.descent)	* 1.2;
	if (y >= 0) y += fst->max_bounds.ascent;
	else {
		for (q = str, i = 0; *q; q++) if (*q == '\n') i ++;
		y += WindowHeight - i*LineHeight - fst->max_bounds.descent;
	}
	for (p = buf, q = str; *q; q ++) {
		if (*q == '%') {
			fmt[0] = *(q++);
			for (i=1, flg=0; i<FMTLEN-1 && !flg && *q;)
			switch (*q) {
				case 'g': k = Generation; flg = 1; break;
				case 's': k = Step; flg = 1; break;
				case 'c': k = ind->size; flg = 1; break;
				case '+': case '-': fmt[i++] = *(q++); break;
				default:
				if (*q >= '0' && *q <= '9') fmt[i++] = *(q++);
				else { strcpy(fmt,"???"); flg = 2; }
			}
			if (!flg) break;
			if (flg == 1) { fmt[i] = 'd'; fmt[i+1] = '\0'; }
			sprintf(p,fmt,k);
			p += strlen(p);
		} else if (*q == '\n') {
			draw_string(cap_gc,buf,fst,x,y,align);
			y += LineHeight;
			p = buf;
		} else { *(p++) = *q; *p = '\0'; }
	}
	if (p > buf) draw_string(cap_gc,buf,fst,x,y,align);
}
static void check_caption()
{
	XFontStruct	*fst;
	char *str; int x, y, align;
	get_caption_string(&str,&fst,&x,&y,&align);
	if (str) draw_caption(ZoomedIndividual,str,fst,x,y,align);
}
static	XImage	*image = NULL;
static	char	*data = NULL;
static	unsigned char	*rgb_data = NULL;
static	void	(*put_pixel_proc)();
static	int	wy = 0, bpp;
extern	int	DeluxePainting;
static	void	store_1byte(x,p) unsigned long x; unsigned char *p;
{ p[0] = x & 0xff; }
static	void	store_lf2(x,p) unsigned long x; unsigned char *p;
{ p[0] = x & 0xff; p[1] = (x >> 8) & 0xff; }
static	void	store_mf2(x,p) unsigned long x; unsigned char *p;
{ p[1] = x & 0xff; p[0] = (x >> 8) & 0xff; }
static	void	store_lf4(x,p) unsigned long x; unsigned char *p;
{ p[0] = x & 0xff; p[1] = (x >> 8) & 0xff;
  p[2] = (x >> 16) & 0xff; p[3] = (x >> 24) & 0xff; }
static	void	store_mf4(x,p) unsigned long x; unsigned char *p;
{ p[3] = x & 0xff; p[2] = (x >> 8) & 0xff;
  p[1] = (x >> 16) & 0xff; p[0] = (x >> 24) & 0xff; }
static	void	set_view_pose_d(c) Cell	*c; {
	double	x, y, z, r;
	set_view_pose(c,&x,&y,&z,&r);
	c->C = x * x + y * y + z * z - r * r;
	c->draw_x = x; c->draw_y = y; c->draw_z = z; c->draw_r = r;
}
Boolean	draw_one_line(cl) XtPointer cl;
{
	int	wy0, nl, wx, r, g, b, flag = 0;
	ViewArrayElm	**varray;
	char	*dp;
	unsigned char	*rgb_dp;
	void	start_exec(), check_run_selection(), draw_cell();
	static	unsigned char	mask[] = {
		0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	if (DrawShape != DrSpDeluxe) goto chain;
	if (wy == 0) {
		if (!ViewArray) {
			ViewArray = (ViewArrayElm **)malloc(
				sizeof(ViewArrayElm *) * ViewArraySize);
			if (!ViewArray) {
				perror("malloc error for ViewArray");
				exit(1);
			}
			memset(ViewArray,0,
				sizeof(ViewArrayElm *)*ViewArraySize);
		} else free_viewarrayelm();
		NObject = 0;
		for_all_cells(ZoomedIndividual,draw_cell);
		for_all_cells(ZoomedIndividual,set_view_array);
		for_all_cells(ZoomedIndividual,set_view_pose_d);
		retro_position(-ViewShift,-ViewHeight,-vz-ViewDistance,
			&VPx,&VPy,&VPz);
		if (!rgb_data) {
			rgb_data = (unsigned char *)malloc(
				WindowWidth * WindowHeight * 3);
			if (!rgb_data)
				show_message("malloc error for rgb_data",NULL);
		}
	}
	for (wy0 = wy; wy < WindowHeight && !flag && wy - wy0 < 10; wy ++) {
		varray = ViewArray + view_array_y(wy) * ViewArrayWidth;
		Sy = (WindowHeight - 1 - wy - CenterY) / scale;
		dp = data + bpp * WindowWidth * wy;
		if (rgb_data) rgb_dp = rgb_data + WindowWidth * wy * 3;
		for (wx = 0; wx < WindowWidth;
			wx ++, dp += bpp, rgb_dp += 3) {
			if (calc_rgb(wx,wy,varray[view_array_x(wx)],&r,&g,&b))
				flag = 1;
			if (rgb_data) {
				rgb_dp[0] = r & 0xff;
				rgb_dp[1] = g & 0xff;
				rgb_dp[2] = b & 0xff;
			}
			(*put_pixel_proc)(rgb_to_pixel(r,g,b),dp);
		}
	}
	XPutImage(dpy,pixmap,gc,image,0,wy0,0,wy0,WindowWidth,wy-wy0);
	XCopyArea(dpy,pixmap,window,gc,0,wy0,WindowWidth,wy-wy0,0,wy0);
	if (wy < WindowHeight) return False;
	check_caption();
	check_dump_frame();
chain:
	wy = DeluxePainting = 0;
	check_run_selection();
	start_exec();
	return True;
}
static	void	create_screen_image()
{
	bpp = (depth <= 8)? 1 : (depth <= 16)? 2 : 4;
	data = (char *)malloc(bpp * WindowWidth * WindowHeight);
	if (!data) {
		show_message("Need more memory.",NULL);
		return;
	}
	image = XCreateImage(dpy,&visual,depth,ZPixmap,0,
		data,WindowWidth,WindowHeight,bpp*8,bpp*WindowWidth);
	if (!image) {
		show_message("Cannot create image data.",NULL);
		return;
	}
	if (bpp == 1) put_pixel_proc = store_1byte;
	else switch (image->byte_order) {
		case LSBFirst:
		put_pixel_proc = (bpp == 2)? store_lf2 : store_lf4;
		break;
		case MSBFirst:
		put_pixel_proc = (bpp == 2)? store_mf2 : store_mf4;
		break;
		default:
		show_message("unknown byte order.",NULL);
		XDestroyImage(image);
		image = NULL;
		return;
	}
}
static	void	draw_ray_tracing(ind)
Individual	*ind;
{
	if (!image) { create_screen_image(); if (!image) return; }
	wy = 0;
	if (DeluxePainting == 0) DeluxePainting = 1;
}

static	void	draw_sphere(c)
Cell	*c;
{
	unsigned long	*pxl;
	int	i, d;
	double	r;
	pxl = pixel + ((int)(c->glu*(NUT_GRADES-.001))*NUT_GRADES
		+ (int)(c->h2o*(NUT_GRADES-.001))) * COLOR_GRADES;
	if ((c->y < 0 && RotateX > 0) || (c->y > 0 && RotateX < 0))
		pxl += NCOLORS;
	if (DrawShape == DrSpSkeleton) {
		XSetForeground(dpy,gc,
			pxl[BlackBackground?COLOR_GRADES-1:COLOR_GRADES/2]);
		if (c->mother) XDrawLine(dpy,pixmap,gc,
			(int)(c->draw_x),(int)(c->draw_y),
			(int)(c->mother->draw_x),(int)(c->mother->draw_y));
		else XDrawPoint(dpy,pixmap,gc,
			(int)(c->draw_x),(int)(c->draw_y));
	} else for (i = 0; i < COLOR_GRADES; i ++) {
		XSetForeground(dpy,gc,pxl[i]);
		r = c->draw_r * radius_rate[i] * 1.1;
		d = r + r;
		XFillArc(dpy,pixmap,gc,
		(int)(c->draw_x - r),(int)(c->draw_y - r),d,d,0,64*360);
	}
}
void	draw_cell(c)
Cell	*c;
{
	double	x, y, z, r, s;
	if (c->step > Step) return;
	if (NObject == MaxObjects) {
		/* fprintf(stderr,"Too many cells: %d\n",NObject); */
		return;
	}
	set_view_pose(c,&x,&y,&z,&r);
 	if (z < 1e-20 || r < 1e-20) return;
	s = scale / z;
	c->draw_x = CenterX + x * s;
	c->draw_y = CenterY - y * s;
	c->draw_z = z;
	c->draw_r = r * s;
	ObjectBuffer[NObject] = c;
	NObject ++;
}
static int comp_zz(a,b)
Cell	**a, **b;
{
	return ((*a)->draw_z < (*b)->draw_z)? 1
		: (((*a)->draw_z > (*b)->draw_z)? -1 : 0);
}

static	void	(*DrawSphereProc)() = draw_sphere;
static	void	draw_individual(ind)
Individual	*ind;
{
	char	buf[16];
	Cell	**p;
	NObject = 0;
	for_all_cells(ind,draw_cell);
	qsort(ObjectBuffer,NObject,sizeof(Cell *),comp_zz);
	for (p = ObjectBuffer; NObject > 0; NObject --, p ++)
		(*DrawSphereProc)(*p);
	if (DrawSphereProc == draw_sphere) {
		sprintf(buf,"%d",ind->size);
		XSetForeground(dpy,gc,fg);
		XDrawString(dpy,pixmap,gc,CenterX-(IndWidth/2)+4,
			CenterY+(IndHeight/2)-4,buf,strlen(buf));
	}
}
Individual	*get_pointed_individual(x,y)
int	x,y;
{
	int	i = (y / IndHeight) * NColumns + (x / IndWidth);
	if (i >= PopulationSize) return NULL;
	return (Population + i);
}

switch_draw_mode(x,y)
int	x,y;
{
	if (ZoomMode) {
		ZoomMode = 0;
		XSetLineAttributes(dpy,gc,0,LineSolid,CapNotLast,JoinMiter);
		redraw_world();
	} else if (ZoomedIndividual = get_pointed_individual(x,y)) {
		ZoomMode = 1;
		XSetLineAttributes(dpy,gc,5,LineSolid,CapNotLast,JoinMiter);
		redraw_world();
	}
}
show_gene(x,y)
int	x,y;
{
	Individual	*ind
	= (ZoomMode)? ZoomedIndividual : get_pointed_individual(x,y);
	if (ind) viewg(ind);
}
static	double	distance(a,b)
Cell	*a, *b;
{
	return (!a || !b)? 0.0 :
	sqrt(pow(a->x-b->x,2.0)+pow(a->y-b->y,2.0)+pow(a->z-b->z,2.0));
}
static	void	print_cell(c)
Cell	*c;
{
	printf("%x(%x):%d,%x(%g,%g,%g),%g\n",
	c,c->mother,c->step,c->state,c->x,c->y,c->z,
	distance(c->mother,c));
}
show_individual(x,y)
int	x,y;
{
	Individual	*ind;
	Cell	*c;
	int	i;
	if (ZoomMode) ind = ZoomedIndividual;
	else if (!(ind = get_pointed_individual(x,y))) return 0;
	printf("--------------- Individual #%d, size = %d -----------\n",
		ind->id,ind->size);
	for (i = 0; i < GeneLen; i++) printf("%08x:",ind->gene[i]);
	putchar('\n');
	for_all_cells(ind,print_cell);
}
static void draw_world()
{
	Individual	*ind;
	int	i, x, y, psize;
	extern	int	GenerationLength;
	if (!DrawPopSize) return;
	sinrx = sin(RotateX); cosrx = cos(RotateX);
	sinry = sin(RotateY); cosry = cos(RotateY);
	sinrz = sin(RotateZ); cosrz = cos(RotateZ);
	scale = ScaleR * IScale * 0.01;
	if (ZoomMode) {
		scale *= WindowWidth;
		IndWidth = WindowWidth; IndHeight = WindowHeight;
		CenterX = WindowWidth / 2; CenterY = WindowHeight / 2;
		if (DrawShape == DrSpDeluxe && DrawSphereProc == draw_sphere)
			draw_ray_tracing(ZoomedIndividual);
		else {
			draw_individual(ZoomedIndividual);
			if (DrawSphereProc == draw_sphere) check_caption();
		}
	} else {
		psize = (DrawPopSize < 0)? PopulationSize : DrawPopSize;
		NColumns = ceil(sqrt((double)psize));
		IndWidth = WindowWidth / NColumns;
		IndHeight = WindowHeight / NColumns;
		scale *= IndWidth;
		XSetForeground(dpy,gc,fg);
		for (i = 1; i < NColumns; i ++) {
			y = i * IndHeight; x = i * IndWidth;
			XDrawLine(dpy,pixmap,gc,0,y,WindowWidth-1,y);
			XDrawLine(dpy,pixmap,gc,x,0,x,WindowHeight-1);
		}
	for (ind = Population, i = 0; i < psize; i ++, ind ++) {
		CenterX = IndWidth / 2 + (i % NColumns) * IndWidth;
		CenterY = IndHeight / 2 + (i / NColumns) * IndHeight;
		draw_individual(ind);
		}
	}
}
redraw_world() {
	extern	void	check_deluxe_draw();
	if (ZoomMode && DrawShape == DrSpDeluxe) {
		DrawSphereProc = draw_sphere; draw_world();
		check_deluxe_draw();
	} else {
		XSetForeground(dpy,gc,bg);
		XFillRectangle(dpy,pixmap,gc,0,0,WindowWidth,WindowHeight);
		DrawSphereProc = draw_sphere; draw_world();
		XCopyArea(dpy,pixmap,window,gc,0,0,
			WindowWidth,WindowHeight,0,0);
		XFlush(dpy);
		check_dump_frame();
	}
}
save_image_p(proc) void (*proc)(); { DrawSphereProc = proc; draw_world(); }

static	char	title[] = "l3m";
ximage_to_xwdhead(i,h)
XImage		*i;
XWDFileHeader	*h;
{
	h->header_size = sizeof(XWDFileHeader) + strlen(title) + 1;
	h->file_version = XWD_FILE_VERSION;
	h->pixmap_format = ZPixmap;
	h->pixmap_depth = i->depth;
	h->pixmap_width = i->width;
	h->pixmap_height = i->height;
	h->xoffset = i->xoffset;
	h->byte_order = i->byte_order;
	h->bitmap_unit = i->bitmap_unit;
	h->bitmap_bit_order = i->bitmap_bit_order;
	h->bitmap_pad = i->bitmap_pad;
	h->bits_per_pixel = i->bits_per_pixel;
	h->bytes_per_line = i->bytes_per_line;
	h->visual_class = visual.class;
	h->red_mask = visual.red_mask;
	h->green_mask = visual.green_mask;
	h->blue_mask = visual.blue_mask;
	h->bits_per_rgb = visual.bits_per_rgb;
	h->colormap_entries = visual.map_entries;
	h->window_width = WindowWidth;
	h->window_height = WindowHeight;
	h->window_x = h->window_y = 0;
	h->window_bdrwidth = 1;
}

save_xwd_image(fd) FILE *fd;
{
	int	i, n, datalen;
	XWDFileHeader	h;
	unsigned char	flg[256], colors[256];
	XColor		xcol[256];
	if (!image) { create_screen_image(); if (!image) return 1; }
	XGetSubImage(dpy,pixmap,0,0,WindowWidth,WindowHeight,
		-1,ZPixmap,image,0,0);
	ximage_to_xwdhead(image,&h);
	datalen = image->bytes_per_line * image->height;
	switch (h.visual_class) {
		case PseudoColor:
		memset(flg,0,256);
		for (i = n = 0; i < datalen; i ++) {
			if (!flg[data[i]]) {
				flg[data[i]] = 1;
				xcol[n].pixel = data[i];
				xcol[n].flags = DoRed | DoGreen | DoBlue;
				colors[data[i]] = n ++;
			}
			data[i] = colors[data[i]];
		}
		h.ncolors = n; break;
		default: h.ncolors = 0;
	}
	fwrite(&h,sizeof(h),1,fd);
	fwrite(title,1,strlen(title)+1,fd);
	if (h.ncolors > 0) {
		XWDColor	xwdcol[256];
		XQueryColors(dpy,cmap,xcol,n);
		for (i = 0; i < n; i ++) {
			xwdcol[i].pixel = i;
			xwdcol[i].red = xcol[i].red;
			xwdcol[i].green = xcol[i].green;
			xwdcol[i].blue = xcol[i].blue;
			xwdcol[i].flags = xcol[i].flags;
		}
		fwrite(xwdcol,sizeof(XWDColor),n,fd);
	}
	fwrite(data,1,datalen,fd);
}

static void put_ps1(fd,c,n) FILE *fd; unsigned char c; int *n;
{
	fprintf(fd,"%02x",c);
	if ((++ *n) >= 36) { fputc('\n',fd); *n = 0; }
}
static void ps_prolog(fd,sw,sh) FILE *fd; int sw, sh;
{
	fputs("gsave\n/R { currentfile exch readhexstring pop } bind def\n",fd);
	fprintf(fd,"%d %d 8 [ %d 0 0 -%d 0 %d ]\n",WindowWidth,WindowHeight,
		WindowWidth,WindowHeight,WindowHeight);
	fprintf(fd,"{%d string R}\n{%d string R}\n{%d string R}\n%d %d scale\n",
		WindowWidth,WindowWidth,WindowWidth,sw,sh);
	fputs("true 3\ncolorimage\n",fd);
}
save_ps_image(fd)
FILE	*fd;
{
	int	y, x, n;
	unsigned char	*pl, *pp;
	if (!rgb_data) { show_message("No RGB data kept.",NULL); return 1; }
	ps_prolog(fd,WindowWidth,WindowHeight);
	for (y=n=0, pl=rgb_data; y<WindowHeight; y++, pl+=WindowWidth*3) {
		for (x=0,pp=pl; x<WindowWidth; x++,pp+=3) put_ps1(fd,*pp,&n);
		for (x=0,pp=pl+1; x<WindowWidth; x++,pp+=3) put_ps1(fd,*pp,&n);
		for (x=0,pp=pl+2; x<WindowWidth; x++,pp+=3) put_ps1(fd,*pp,&n);
	}
	if (n > 0) fputc('\n',fd);
	return 0;
}
static	int	SaveWidth, SaveHeight, SaveCenterX, SaveCenterY;
static	double	SaveScale;
static	unsigned char	*PrintBuf = NULL;
static	int	PrintBufSize = 0;
static	FILE	*PrintFD;
print_ps_image(pout,dpi)
FILE	*pout;
int	dpi;
{
	int	s;
	SaveWidth = WindowWidth; SaveHeight = WindowHeight;
	WindowWidth = 640 * dpi / 72;
	WindowHeight = 480 * dpi / 72;
	if (SaveWidth == WindowWidth && SaveHeight == WindowHeight) {
		save_ps_image(pout);
		print_close(pout);
		return 1;
	}
	s = WindowWidth * 3;
	if (PrintBufSize < s) {
		PrintBuf = (unsigned char *)malloc(s);
		if (!PrintBuf) {
			show_message("malloc error for PS print buffer",NULL);
			return 1;
		}
		PrintBufSize = s;
	}
	progress_meter_start();
	SaveCenterX = CenterX; SaveCenterY = CenterY; SaveScale = scale;
	CenterX = WindowWidth / 2;
	CenterY = WindowHeight / 2;
	scale = ScaleR * IScale * 0.01 * WindowWidth;
	ps_prolog(pout,640,480);
	PrintFD = pout;
	DrawShape = DrSpGrowingSpheres;
	wy = 0;
	return 0;
}
Boolean print_one_line(cl) XtPointer cl;
{
	int	x, r, g, b;
	static	int	n;
	ViewArrayElm	**varray;
	if (wy == 0) n = 0;
	varray = ViewArray + view_array_y(wy) * ViewArrayWidth;
	Sy = (WindowHeight - 1 - wy - CenterY) / scale;
	for (x = 0; x < WindowWidth; x ++) {
		calc_rgb(x,wy,varray[view_array_x(x)],&r,&g,&b);
		PrintBuf[x] = r & 0xff;
		PrintBuf[WindowWidth + x] = g & 0xff;
		PrintBuf[WindowWidth*2 + x] = b & 0xff;
	}
	for (x = 0; x < WindowWidth*3; x ++) put_ps1(PrintFD,PrintBuf[x],&n);
	progress_meter_set(wy,WindowHeight);
	if ((++ wy) < WindowHeight) return False;
	print_close(PrintFD);
	WindowWidth = SaveWidth; WindowHeight = SaveHeight;
	CenterX = SaveCenterX; CenterY = SaveCenterY; scale = SaveScale;
	progress_meter_close();
	DrawShape = DrSpDeluxe;
	return True;
}
print_cancel() { wy = WindowHeight - 1; }
static unsigned long	get_pixel_from_image(x,y) int x,y;
{
	int	bpp = image->bits_per_pixel / 8;
	unsigned char	*p = (unsigned char *)data
		+ image->bytes_per_line * y + bpp * x;
	switch (bpp) {
		case 2: return (image->byte_order == LSBFirst)?
			(p[1] << 8) | p[0] : (p[0] << 8) | p[1];
		case 4: return (image->byte_order == LSBFirst)?
			(p[3] << 24) | (p[2] << 16) | (p[1] << 8) | p[0] :
			(p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
		default: return (unsigned long)(*p);
	}
}
static unsigned int	mask_shift(pxl,m,s) unsigned long pxl, m; int s;
{
	unsigned int	v = pxl & m;
	return (s > 0)? (v >> s) : (v << -s);
}
static void	pixel_to_rgb(pxl,rgb) unsigned long pxl; unsigned char *rgb;
{
	if (visual.class != PseudoColor) {
		rgb[0] = mask_shift(pxl,visual.red_mask,red_shift);
		rgb[1] = mask_shift(pxl,visual.green_mask,green_shift);
		rgb[2] = mask_shift(pxl,visual.blue_mask,blue_shift);
	}
}
static int	write_rgb_from_image(fd)
FILE	*fd;
{
	static	unsigned char	*buf = NULL;
	int	x, y, bpp;
	unsigned char	*p;
	if (!image) { create_screen_image(); if (!image) return 1; }
	XGetSubImage(dpy,pixmap,0,0,WindowWidth,WindowHeight,
		-1,ZPixmap,image,0,0);
	if (!buf) {
		buf = (unsigned char *)malloc(WindowWidth*3);
		if (!buf) {
			show_message("malloc error",NULL);
			return 0;
		}
	}
	for (y = 0; y < WindowHeight; y ++) {
		for (x = 0, p = buf; x < WindowWidth; x ++, p += 3)
			pixel_to_rgb(get_pixel_from_image(x,y),p);
		fwrite(buf,1,WindowWidth*3,fd);
	}
}
save_one_frame(fd)
FILE	*fd;
{
	return (DrawShape == DrSpDeluxe && rgb_data)?
		fwrite(rgb_data,1,WindowWidth*WindowHeight*3,fd) :
		write_rgb_from_image(fd);
}
